@@ -17,7 +17,6 @@
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
-#include <termios.h>
 #include <sys/types.h>
 #include <sys/time.h>
 #include <sys/socket.h>
@@ -32,6 +31,11 @@
 #include "controller.h"
 #include "selector.h"
 #include "dataxfer.h"
+#include "utils.h"
+#include "telnet.h"
+
+
+extern selector_t *ser2net_sel;
 
 /** BASED ON sshd.c FROM openssh.com */
 #ifdef HAVE_TCPD_H
@@ -43,7 +47,6 @@ static char *progname = "ser2net-control";
 
 /* This file holds the code that runs the control port. */
 
-static int port;	/* The TCP port for the control port. */
 static int acceptfd;	/* The file descriptor for the accept port. */
 
 static int max_controller_ports = 4;	/* How many control connections
@@ -63,10 +66,6 @@ typedef struct controller_info {
     struct sockaddr_in remote;		/* The socket address of who
 					   is connected to this port. */
 
-    unsigned char telnet_cmd[3];	/* An incoming telnet command. */
-    int  telnet_cmd_pos;		/* The current position in the
-					   telnet command buffer. */
-
     unsigned char inbuf[INBUF_SIZE+1];	/* Buffer to receive command on. */
     int  inbuf_count;			/* The number of bytes currently
 					   in the inbuf. */
@@ -88,16 +87,28 @@ typedef struct controller_info {
 
     struct controller_info *next;	/* Used to keep these items in
 					   a linked list. */
+
+    /* Data used by the telnet processing. */
+    telnet_data_t tn_data;
 } controller_info_t;
 
 /* List of current control connections. */
 controller_info_t *controllers = NULL;
 
 /* Used to initialize the telnet session. */
-static char telnet_init[] = {
-    0xff, 0xfb, 0x03,  /* command WILL SUPPRESS GO AHEAD */
-    0xff, 0xfb, 0x01,  /* command WILL ECHO */
-    0xff, 0xfe, 0x01   /* command DON'T ECHO */
+static char telnet_init_seq[] = {
+    TN_IAC, TN_WILL, TN_OPT_SUPPRESS_GO_AHEAD,
+    TN_IAC, TN_WILL, TN_OPT_ECHO,
+    TN_IAC, TN_DONT, TN_OPT_ECHO,
+};
+
+static struct telnet_cmd telnet_cmds[] = 
+{
+    /*                        I will,  I do,  sent will, sent do */
+    { TN_OPT_SUPPRESS_GO_AHEAD,	   0,     1,          1,       0, },
+    { TN_OPT_ECHO,		   0,     1,          1,       1, },
+    { TN_OPT_BINARY_TRANSMISSION,  1,     1,          0,       1, },
+    { 255 }
 };
 
 /* Shut down a control connection and remove it from the list of
@@ -113,7 +124,7 @@ shutdown_controller(controller_info_t *cntlr)
 	cntlr->monitor_port_id = NULL;
     }
 
-    clear_fd_handlers(cntlr->tcpfd);
+    sel_clear_fd_handlers(ser2net_sel, cntlr->tcpfd);
     close(cntlr->tcpfd);
     if (cntlr->outbuf != NULL) {
 	free(cntlr->outbuf);
@@ -153,15 +164,14 @@ controller_output(struct controller_info *cntlr,
 	/* Already outputting data, just add more onto it. */
 	int  new_size = cntlr->outbuf_count + count;
 
-	if (new_size < cntlr->outbufsize) {
+	if (new_size <= cntlr->outbufsize) {
 	    /* It will fit into the current buffer, just move things
 	       around and append it. */
 	    int i;
 
 	    if (cntlr->outbuf_pos > 0) {
 		for (i=0; i<cntlr->outbuf_count; i++) {
-		    cntlr->outbuf[i] = cntlr->outbuf[cntlr->outbuf_pos];
-		    (cntlr->outbuf_pos)++;
+		    cntlr->outbuf[i] = cntlr->outbuf[cntlr->outbuf_pos + i];
 		}
 	    }
 	    memcpy(&(cntlr->outbuf[cntlr->outbuf_count]), data, count);
@@ -187,27 +197,30 @@ controller_output(struct controller_info *cntlr,
 	    memcpy(newbuf+cntlr->outbuf_count, data, count);
 	    free(cntlr->outbuf);
 	    cntlr->outbuf = newbuf;
-	    cntlr->outbuf_pos = 0;
 	}
+	cntlr->outbuf_pos = 0;
 	cntlr->outbuf_count += count;
     } else {
 	/* We are starting a new buffer, just get it. */
 	char *newbuf;
+	int  new_size = ((count / 1024) * 1024) + 1024;
 
-	newbuf = malloc(1024);
+	newbuf = malloc(new_size);
 	if (newbuf == NULL) {
 	    /* Out of memory, just ignore thre request */
 	    return;
 	}
 	
-	cntlr->outbufsize = 1024;
+	cntlr->outbufsize = new_size;
 
 	memcpy(newbuf, data, count);
 	cntlr->outbuf = newbuf;
 	cntlr->outbuf_pos = 0;
 	cntlr->outbuf_count = count;
-	set_fd_read_handler(cntlr->tcpfd, FD_HANDLER_DISABLED);
-	set_fd_write_handler(cntlr->tcpfd, FD_HANDLER_ENABLED);
+	sel_set_fd_read_handler(ser2net_sel, cntlr->tcpfd,
+				SEL_FD_HANDLER_DISABLED);
+	sel_set_fd_write_handler(ser2net_sel, cntlr->tcpfd,
+				 SEL_FD_HANDLER_ENABLED);
     }
 }
 
@@ -218,9 +231,20 @@ controller_write(struct controller_info *cntlr, char *data, int count)
     write(cntlr->tcpfd, data, count);
 }
 
-/* Called when a telnet command is received complete. */
+static void
+telnet_output_ready(void *cb_data)
+{
+    struct controller_info *cntlr = cb_data;
+
+    sel_set_fd_read_handler(ser2net_sel, cntlr->tcpfd,
+			    SEL_FD_HANDLER_DISABLED);
+    sel_set_fd_write_handler(ser2net_sel, cntlr->tcpfd,
+			     SEL_FD_HANDLER_ENABLED);
+}
+
+/* Called when a telnet command is received. */
 void
-process_telnet_command(controller_info_t *cntlr)
+telnet_cmd_handler(void *cb_data, unsigned char cmd)
 {
     /* These are ignored for now. */
 }
@@ -241,6 +265,8 @@ static char *help_str =
 "disconnect <tcp port> - disconnect the tcp connection on the port.\n\r"
 "showport [<tcp port>] - Show information about a port. If no port is\n\r"
 "       given, all ports are displayed.\n\r"
+"showshortport [<tcp port>] - Show information about a port in a one-line\n\r"
+"       format. If no port is given, all ports are displayed.\n\r"
 "setporttimeout <tcp port> <timeout> - Set the amount of time in seconds\n\r"
 "       before the port connection will be shut down if no activity\n\r"
 "       has been seen on the port.\n\r"
@@ -248,7 +274,7 @@ static char *help_str =
 "       the device configuration in the ser2net.conf file.  Valid options\n\r"
 "       are: 300, 1200, 2400, 4800, 9600, 19200, 38400, 115200, EVEN, ODD\n\r"
 "       NONE, 1STOPBIT, 2STOPBITS, 7DATABITS, 8DATABITS, LOCAL (ignore\n\r"
-"       model control).\n\r"
+"       modem control), [-]RTSCTS, [-]XONXOFF.\n\r"
 "       Note that these will not change until the port is disconnected\n\r"
 "       and connected again.\n\r"
 "setportcontrol <tcp port> <controls>\n\r"
@@ -259,6 +285,7 @@ static char *help_str =
 "       Valid states are:\n\r"
 "         off - The TCP port is shut down\n\r"
 "         raw - The TCP port is up and all I/O is transferred\n\r"
+"         rawlp - The TCP port is up and the input is transferred to dev\n\r"
 "         telnet - The TCP port is up and the telnet negotiation protocol\n\r"
 "                  runs on the port.\n\r";
 
@@ -277,6 +304,9 @@ process_input_line(controller_info_t *cntlr)
     } else if (strcmp(tok, "exit") == 0) {
 	shutdown_controller(cntlr);
 	return; /* We don't want a prompt any more. */
+    } else if (strcmp(tok, "quit") == 0) {
+	shutdown_controller(cntlr);
+	return; /* We don't want a prompt any more. */
     } else if (strcmp(tok, "help") == 0) {
 	controller_output(cntlr, help_str, strlen(help_str));
     } else if (strcmp(tok, "version") == 0) {
@@ -288,6 +318,9 @@ process_input_line(controller_info_t *cntlr)
     } else if (strcmp(tok, "showport") == 0) {
 	tok = strtok_r(NULL, " \t", &strtok_data);
 	showports(cntlr, tok);
+    } else if (strcmp(tok, "showshortport") == 0) {
+	tok = strtok_r(NULL, " \t", &strtok_data);
+	showshortports(cntlr, tok);
     } else if (strcmp(tok, "monitor") == 0) {
 	tok = strtok_r(NULL, " \t", &strtok_data);
 	if (tok == NULL) {
@@ -450,42 +483,16 @@ handle_tcp_fd_read(int fd, void *data)
 	return;
     }
     read_start = cntlr->inbuf_count;
+    read_count = process_telnet_data
+	(cntlr->inbuf+read_start, read_count, &cntlr->tn_data);
+    if (cntlr->tn_data.error) {
+	shutdown_controller(cntlr);
+	return;
+    }
     cntlr->inbuf_count += read_count;
-
+    
     for (i=read_start; i<cntlr->inbuf_count; i++) {
-	if (cntlr->telnet_cmd_pos != 0) {
-	    if ((cntlr->telnet_cmd_pos == 1)
-		&& (cntlr->inbuf[i] == 255))
-	    {
-		/* Two IACs in a row causes one IAC to be sent, so
-		   just let this one go through. */
-		continue;
-	    }
-
-	    /* In the middle of a telnet command. */
-	    cntlr->telnet_cmd[cntlr->telnet_cmd_pos] = cntlr->inbuf[i];
-	    cntlr->telnet_cmd_pos++;
-
-	    i = remove_chars(cntlr, i, 1);
-
-	    if ((cntlr->telnet_cmd_pos == 2)
-		&& (cntlr->telnet_cmd[1] <= 250))
-	    {
-		/* These are two byte commands, so we have
-		   everything we need to handle the command. */
-		process_telnet_command(cntlr);
-		cntlr->telnet_cmd_pos = 0;
-	    } else if (cntlr->telnet_cmd_pos == 3) {
-		/* We are done with the telnet command. */
-		process_telnet_command(cntlr);
-		cntlr->telnet_cmd_pos = 0;
-	    }
-	} else if (cntlr->inbuf[i] == 0xff) {
-	    /* Got a telnet command start. */
-	    cntlr->telnet_cmd[cntlr->telnet_cmd_pos] = cntlr->inbuf[i];
-	    cntlr->telnet_cmd_pos = 1;
-	    i = remove_chars(cntlr, i, 1);
-	} else if (cntlr->inbuf[i] == 0x0) {
+	if (cntlr->inbuf[i] == 0x0) {
 	    /* Ignore nulls. */
 	    i = remove_chars(cntlr, i, 1);
 	} else if (cntlr->inbuf[i] == '\n') {
@@ -533,8 +540,42 @@ static void
 handle_tcp_fd_write(int fd, void *data)
 {
     controller_info_t *cntlr = (controller_info_t *) data;
+    telnet_data_t *td = &cntlr->tn_data;
     int write_count;
 
+    if (td->out_telnet_cmd_size > 0) {
+	write_count = write(cntlr->tcpfd,
+			    &(td->out_telnet_cmd[0]),
+			    td->out_telnet_cmd_size);
+	if (write_count == -1) {
+	    if (errno == EINTR) {
+		/* EINTR means we were interrupted, just retry by returning. */
+		return;
+	    }
+
+	    if (errno == EAGAIN) {
+		/* This again was due to O_NONBLOCK, just ignore it. */
+	    } else if (errno == EPIPE) {
+		shutdown_controller(cntlr);
+	    } else {
+		/* Some other bad error. */
+		syslog(LOG_ERR, "The tcp write for controller had error: %m");
+		shutdown_controller(cntlr);
+	    }
+	} else {
+	    int i, j;
+
+	    /* Copy the remaining data. */
+	    for (j=0, i=write_count; i<td->out_telnet_cmd_size; i++, j++)
+		td->out_telnet_cmd[j] = td->out_telnet_cmd[i];
+	    td->out_telnet_cmd_size -= write_count;
+	    if (td->out_telnet_cmd_size != 0)
+		/* If we have more telnet command data to send, don't
+		   send any real data. */
+		return;
+	}
+    }
+
     write_count = write(cntlr->tcpfd,
 			&(cntlr->outbuf[cntlr->outbuf_pos]),
 			cntlr->outbuf_count);
@@ -557,8 +598,10 @@ handle_tcp_fd_write(int fd, void *data)
 	    /* We are done writing, turn the reader back on. */
 	    free(cntlr->outbuf);
 	    cntlr->outbuf = NULL;
-	    set_fd_read_handler(cntlr->tcpfd, FD_HANDLER_ENABLED);
-	    set_fd_write_handler(cntlr->tcpfd, FD_HANDLER_DISABLED);
+	    sel_set_fd_read_handler(ser2net_sel, cntlr->tcpfd,
+				    SEL_FD_HANDLER_ENABLED);
+	    sel_set_fd_write_handler(ser2net_sel, cntlr->tcpfd,
+				     SEL_FD_HANDLER_DISABLED);
 	}
     }
 }
@@ -581,32 +624,21 @@ handle_accept_port_read(int fd, void *data)
     socklen_t         len;
     char              *err = NULL;
 
-    cntlr = malloc(sizeof(*cntlr));
-    if (cntlr == NULL) {
-	err = "Could not allocate controller port\n\r";
-    }
-
     if (num_controller_ports >= max_controller_ports) {
 	err = "Too many controller ports\n\r";
-    }
-
-    if (err != NULL) {
-	/* We have a problem so refuse this one. */
-	struct sockaddr_in dummy_sockaddr;
-	socklen_t len = sizeof(dummy_sockaddr);
-	int new_fd = accept(fd, &dummy_sockaddr, &len);
-
-	if (new_fd != -1) {
-	    write(new_fd, err, strlen(err));
-	    close(new_fd);
+	goto errout2;
+    } else {
+	cntlr = malloc(sizeof(*cntlr));
+	if (cntlr == NULL) {
+	    err = "Could not allocate controller port\n\r";
+	    goto errout2;
 	}
-	return;
     }
 
     /* From here on, errors must go to errout. */
 
     len = sizeof(cntlr->remote);
-    cntlr->tcpfd = accept(fd, &(cntlr->remote), &len);
+    cntlr->tcpfd = accept(fd, (struct sockaddr *) &(cntlr->remote), &len);
     if (cntlr->tcpfd == -1) {
 	syslog(LOG_ERR, "Could not accept on controller port: %m");
 	goto errout;
@@ -623,7 +655,7 @@ handle_accept_port_read(int fd, void *data)
 	    char *err = "Access denied\n\r";
 	    write(cntlr->tcpfd, err, strlen(err));
 	    close(cntlr->tcpfd);
-	    return;
+	    goto errout;
 	}
     }
 #endif /* HAVE_TCPD_H */
@@ -635,14 +667,15 @@ handle_accept_port_read(int fd, void *data)
     }
 
     cntlr->inbuf_count = 0;
-    cntlr->telnet_cmd_pos = 0;
     cntlr->outbuf = NULL;
+    cntlr->monitor_port_id = NULL;
 
-    set_fd_handlers(cntlr->tcpfd,
-		    cntlr,
-		    handle_tcp_fd_read,
-		    handle_tcp_fd_write,
-		    handle_tcp_fd_except);
+    sel_set_fd_handlers(ser2net_sel,
+			cntlr->tcpfd,
+			cntlr,
+			handle_tcp_fd_read,
+			handle_tcp_fd_write,
+			handle_tcp_fd_except);
 
     cntlr->next = controllers;
     controllers = cntlr;
@@ -651,7 +684,10 @@ handle_accept_port_read(int fd, void *data)
        putting the data in the dev to tcp buffer and turning
        the tcp write selector on. */
 
-    controller_output(cntlr, telnet_init, sizeof(telnet_init));
+    telnet_init(&cntlr->tn_data, cntlr, telnet_output_ready,
+		telnet_cmd_handler,
+		telnet_cmds,
+		telnet_init_seq, sizeof(telnet_init_seq));
     controller_output(cntlr, prompt, strlen(prompt));
 
     num_controller_ports++;
@@ -661,16 +697,31 @@ handle_accept_port_read(int fd, void *data)
 errout:
     free(cntlr);
     return;
+
+errout2:
+    {
+	/* We have a problem so refuse this one. */
+	struct sockaddr_in dummy_sockaddr;
+	socklen_t len = sizeof(dummy_sockaddr);
+	int new_fd = accept(fd, (struct sockaddr *) &dummy_sockaddr, &len);
+
+	if (new_fd != -1) {
+	    write(new_fd, err, strlen(err));
+	    close(new_fd);
+	}
+    }
 }
 
 /* Set up the controller port to accept connections. */
-void
-controller_init(int controller_port)
+int
+controller_init(char *controller_port)
 {
     struct sockaddr_in sock;
     int    optval = 1;
 
-    port = controller_port;
+    if (scan_tcp_port(controller_port, &sock) == -1) {
+	return -1;
+    }
 
     acceptfd = socket(PF_INET, SOCK_STREAM, 0);
     if (acceptfd == -1) {
@@ -694,10 +745,7 @@ controller_init(int controller_port)
 	exit(1);
     }
 
-    sock.sin_family = AF_INET;
-    sock.sin_port = htons(port);
-    sock.sin_addr.s_addr = INADDR_ANY;
-    if (bind(acceptfd, &sock, sizeof(sock)) == -1) {
+    if (bind(acceptfd, (struct sockaddr *) &sock, sizeof(sock)) == -1) {
 	close(acceptfd);
 	syslog(LOG_ERR, "Unable to bind TCP port: %m");
 	exit(1);
@@ -709,10 +757,12 @@ controller_init(int controller_port)
 	exit(1);
     }
 
-    set_fd_handlers(acceptfd,
-		    NULL,
-		    handle_accept_port_read,
-		    NULL,
-		    NULL);
-    set_fd_read_handler(acceptfd, FD_HANDLER_ENABLED);
+    sel_set_fd_handlers(ser2net_sel,
+			acceptfd,
+			NULL,
+			handle_accept_port_read,
+			NULL,
+			NULL);
+    sel_set_fd_read_handler(ser2net_sel, acceptfd, SEL_FD_HANDLER_ENABLED);
+    return 0;
 }
