@@ -163,7 +163,7 @@ typedef struct port_info
     int            dev_to_tcp_state;		/* State of transferring
 						   data from the device to
                                                    the TCP port. */
-    char           dev_to_tcp_buf[PORT_BUFSIZE]; /* Buffer used for
+    unsigned char  dev_to_tcp_buf[PORT_BUFSIZE]; /* Buffer used for
                                                     device to TCP
                                                     transfers. */
     int            dev_to_tcp_buf_start;	/* The first byte in
@@ -405,7 +405,14 @@ handle_dev_fd_read(int fd, void *data)
     int write_count;
 
     port->dev_to_tcp_buf_start = 0;
-    port->dev_to_tcp_buf_count = read(fd, port->dev_to_tcp_buf, PORT_BUFSIZE);
+    if (port->enabled == PORT_TELNET) {
+	/* Leave room for IACs. */
+	port->dev_to_tcp_buf_count = read(fd, port->dev_to_tcp_buf,
+					  PORT_BUFSIZE/2);
+    } else {
+	port->dev_to_tcp_buf_count = read(fd, port->dev_to_tcp_buf,
+					  PORT_BUFSIZE);
+    }
 
     if (port->dev_to_tcp_buf_count < 0) {
 	/* Got an error on the read, shut down the port. */
@@ -419,9 +426,27 @@ handle_dev_fd_read(int fd, void *data)
     }
 
     port->dev_bytes_received += port->dev_to_tcp_buf_count;
+
+    if (port->enabled == PORT_TELNET) {
+	int i, j;
+
+	/* Double the IACs on a telnet stream.  This will fit because
+	   we only use half the buffer for telnet connections. */
+	for (i=0; i<port->dev_to_tcp_buf_count; i++) {
+	    if (port->dev_to_tcp_buf[i] == 255) {
+		for (j=port->dev_to_tcp_buf_count; j>i; j--)
+		    port->dev_to_tcp_buf[j+1] = port->dev_to_tcp_buf[j];
+		port->dev_to_tcp_buf_count++;
+		i++;
+		port->dev_to_tcp_buf[i] = 255;
+	    }
+	}
+    }
+
     write_count = write(port->tcpfd,
 			port->dev_to_tcp_buf,
 			port->dev_to_tcp_buf_count);
+
     if (port->dev_monitor != NULL) {
 	controller_write(port->dev_monitor,
 			 port->dev_to_tcp_buf,
@@ -443,11 +468,13 @@ handle_dev_fd_read(int fd, void *data)
 	    port->dev_to_tcp_state = PORT_WAITING_OUTPUT_CLEAR;
 	} else if (errno == EPIPE) {
 	    shutdown_port(port, "EPIPE");
+	    return;
 	} else {
 	    /* Some other bad error. */
 	    syslog(LOG_ERR, "The tcp write for port %s had error: %m",
 		   port->portname);
 	    shutdown_port(port, "tcp write error");
+	    return;
 	}
     } else {
 	port->tcp_bytes_sent += write_count;
@@ -492,6 +519,7 @@ handle_dev_fd_write(int fd, void *data)
 	    syslog(LOG_ERR, "The dev write for port %s had error: %m",
 		   port->portname);
 	    shutdown_port(port, "dev write error");
+	    return;
 	}
     } else {
 	port->dev_bytes_sent += write_count;
@@ -588,6 +616,7 @@ handle_tcp_fd_read(int fd, void *data)
 	    syslog(LOG_ERR, "The dev write for port %s had error: %m",
 		   port->portname);
 	    shutdown_port(port, "dev write error");
+	    return;
 	}
     } else {
 	port->dev_bytes_sent += write_count;
@@ -631,11 +660,13 @@ handle_tcp_fd_write(int fd, void *data)
 		/* This again was due to O_NONBLOCK, just ignore it. */
 	    } else if (errno == EPIPE) {
 		shutdown_port(port, "EPIPE");
+		return;
 	    } else {
 		/* Some other bad error. */
 		syslog(LOG_ERR, "The tcp write for port %s had error: %m",
 		       port->portname);
 		shutdown_port(port, "tcp write error");
+		return;
 	    }
 	} else {
 	    int i, j;
@@ -664,11 +695,13 @@ handle_tcp_fd_write(int fd, void *data)
 	    /* This again was due to O_NONBLOCK, just ignore it. */
 	} else if (errno == EPIPE) {
 	    shutdown_port(port, "EPIPE");
+	    return;
 	} else {
 	    /* Some other bad error. */
 	    syslog(LOG_ERR, "The tcp write for port %s had error: %m",
 		   port->portname);
 	    shutdown_port(port, "tcp write error");
+	    return;
 	}
     } else {
 	port->tcp_bytes_sent += write_count;
@@ -854,46 +887,17 @@ display_banner(port_info_t *port)
     }
 }
 
-/* A connection request has come in on a port. */
-static void
-handle_accept_port_read(int fd, void *data)
+/* Called to set up a new connection's file descriptor. */
+static int
+setup_tcp_port(port_info_t *port)
 {
-    port_info_t *port = (port_info_t *) data;
-    socklen_t len;
-    char *err = NULL;
     int options;
     struct timeval then;
 
-    if (port->tcp_to_dev_state != PORT_UNCONNECTED) {
-	err = "Port already in use\n\r";
-    } else if (is_device_already_inuse(port)) {
-	err = "Port's device already in use\n\r";
-    }
-
-    if (err != NULL) {
-	struct sockaddr_in dummy_sockaddr;
-	socklen_t len = sizeof(dummy_sockaddr);
-	int new_fd = accept(fd, (struct sockaddr *) &dummy_sockaddr, &len);
-
-	if (new_fd != -1) {
-	    write(new_fd, err, strlen(err));
-	    close(new_fd);
-	}
-	return;
-    }
-
-    len = sizeof(port->remote);
-
-    port->tcpfd = accept(fd, (struct sockaddr *) &(port->remote), &len);
-    if (port->tcpfd == -1) {
-	syslog(LOG_ERR, "Could not accept on port %s: %m", port->portname);
-	return;
-    }
-
     if (fcntl(port->tcpfd, F_SETFL, O_NONBLOCK) == -1) {
 	close(port->tcpfd);
 	syslog(LOG_ERR, "Could not fcntl the tcp port %s: %m", port->portname);
-	return;
+	return -1;
     }
 
 #ifdef HAVE_TCPD_H
@@ -907,7 +911,7 @@ handle_accept_port_read(int fd, void *data)
 	    char *err = "Access denied\n\r";
 	    write(port->tcpfd, err, strlen(err));
 	    close(port->tcpfd);
-	    return;
+	    return -1;
 	}
     }
 #endif /* HAVE_TCPD_H */
@@ -923,14 +927,14 @@ handle_accept_port_read(int fd, void *data)
 	    err = "Port already in use by another process\n\r";
 	    write(port->tcpfd, err, strlen(err));
 	    close(port->tcpfd);
-	    return;
+	    return -1;
 	} else if (rv < 0) {
 	    char *err;
 
 	    err = "Error creating port lock file\n\r";
 	    write(port->tcpfd, err, strlen(err));
 	    close(port->tcpfd);
-	    return;
+	    return -1;
 	}
     }
 #endif /* USE_UUCP_LOCKING */
@@ -949,7 +953,10 @@ handle_accept_port_read(int fd, void *data)
 	syslog(LOG_ERR, "Could not open device %s for port %s: %m",
 	       port->devname,
 	       port->portname);
-	return;
+#ifdef USE_UUCP_LOCKING
+	uucp_rm_lock(port->devname);
+#endif /* USE_UUCP_LOCKING */
+	return -1;
     }
 
     if (port->enabled != PORT_RAWLP &&
@@ -959,17 +966,23 @@ handle_accept_port_read(int fd, void *data)
 	syslog(LOG_ERR, "Could not set up device %s for port %s: %m",
 	       port->devname,
 	       port->portname);
-	return;
+#ifdef USE_UUCP_LOCKING
+	uucp_rm_lock(port->devname);
+#endif /* USE_UUCP_LOCKING */
+	return -1;
     }
 
-    /* Turn of BREAK. */
+    /* Turn off BREAK. */
     if (ioctl(port->devfd, TIOCCBRK) == -1) {
 	close(port->tcpfd);
 	close(port->devfd);
 	syslog(LOG_ERR, "Could not turn off break for device %s port %s: %m",
 	       port->devname,
 	       port->portname);
-	return;
+#ifdef USE_UUCP_LOCKING
+	uucp_rm_lock(port->devname);
+#endif /* USE_UUCP_LOCKING */
+	return -1;
     }
     port->is_2217 = 0;
     port->break_set = 0;
@@ -1020,6 +1033,44 @@ handle_accept_port_read(int fd, void *data)
     sel_start_timer(port->timer, &then);
 
     reset_timer(port);
+    return 0;
+}
+
+/* A connection request has come in on a port. */
+static void
+handle_accept_port_read(int fd, void *data)
+{
+    port_info_t *port = (port_info_t *) data;
+    socklen_t len;
+    char *err = NULL;
+
+    if (port->tcp_to_dev_state != PORT_UNCONNECTED) {
+	err = "Port already in use\n\r";
+    } else if (is_device_already_inuse(port)) {
+	err = "Port's device already in use\n\r";
+    }
+
+    if (err != NULL) {
+	struct sockaddr_in dummy_sockaddr;
+	socklen_t len = sizeof(dummy_sockaddr);
+	int new_fd = accept(fd, (struct sockaddr *) &dummy_sockaddr, &len);
+
+	if (new_fd != -1) {
+	    write(new_fd, err, strlen(err));
+	    close(new_fd);
+	}
+	return;
+    }
+
+    len = sizeof(port->remote);
+
+    port->tcpfd = accept(fd, (struct sockaddr *) &(port->remote), &len);
+    if (port->tcpfd == -1) {
+	syslog(LOG_ERR, "Could not accept on port %s: %m", port->portname);
+	return;
+    }
+
+    setup_tcp_port(port);
 }
 
 /* Start monitoring for connections on a specific port. */
@@ -1027,7 +1078,22 @@ static char *
 startup_port(port_info_t *port)
 {
     int optval = 1;
-    
+
+    if (port->tcpport.sin_port == 0) {
+	/* A zero port means use stdin/stdout */
+	if (is_device_already_inuse(port)) {
+	    char *err = "Port's device already in use\n\r";
+	    write(0, err, strlen(err));
+	    exit(1);
+	} else {
+	    port->acceptfd = -1;
+	    port->tcpfd = 0; /* stdin */
+	    if (setup_tcp_port(port) == -1)
+		exit(1);
+	}
+	return NULL;
+    }
+
     port->acceptfd = socket(PF_INET, SOCK_STREAM, 0);
     if (port->acceptfd == -1) {
 	return "Unable to create TCP socket";
@@ -1140,6 +1206,13 @@ shutdown_port(port_info_t *port, char *reason)
     port->dev_bytes_received = 0;
     port->dev_bytes_sent = 0;
 
+    if (port->tcpport.sin_port == 0) {
+	/* This was a zero port (for stdin/stdout), this is only
+	   allowed with one port at a time, and we shut down when it
+	   closes. */
+	exit(0);
+    }
+
     /* If the port has been disabled, then delete it.  Check this before
        the new config so the port will be deleted properly and not
        reconfigured on a reconfig. */
@@ -1266,7 +1339,7 @@ portconfig(char *portnum,
 	return "Could not allocate timer data";
     }
 
-    /* Error from here on out must goto errout. */
+    /* Errors from here on out must goto errout. */
     init_port_data(new_port);
 
     new_port->portname = malloc(strlen(portnum)+1);
@@ -1870,6 +1943,12 @@ disconnect_port(struct controller_info *cntlr,
 	controller_output(cntlr, portspec, strlen(portspec));
  	controller_output(cntlr, "\n\r", 2);
  	return;
+    } else if (port->tcp_to_dev_state == PORT_UNCONNECTED) {
+	char *err = "Port not connected: ";
+ 	controller_output(cntlr, err, strlen(err));
+	controller_output(cntlr, portspec, strlen(portspec));
+ 	controller_output(cntlr, "\n\r", 2);
+ 	return;
     }
  
     shutdown_port(port, "disconnect");
@@ -2219,6 +2298,7 @@ com_port_handler(void *cb_data, unsigned char *option, int len)
 		val = 8;
 	    else
 		val = 9;
+	    break;
 
 	/* RTS handling */
 	case 11:
@@ -2239,6 +2319,7 @@ com_port_handler(void *cb_data, unsigned char *option, int len)
 		val = 11;
 	    else
 		val = 12;
+	    break;
 	}
 
 	outopt[0] = 44;
@@ -2249,22 +2330,36 @@ com_port_handler(void *cb_data, unsigned char *option, int len)
 
     case 8: /* FLOWCONTROL-SUSPEND */
 	tcflow(port->devfd, TCIOFF);
+	outopt[0] = 44;
+	outopt[1] = 108;
+	telnet_send_option(&port->tn_data, outopt, 2);
 	break;
 
     case 9: /* FLOWCONTROL-RESUME */
 	tcflow(port->devfd, TCION);
+	outopt[0] = 44;
+	outopt[1] = 109;
+	telnet_send_option(&port->tn_data, outopt, 2);
 	break;
 
     case 10: /* SET-LINESTATE-MASK */
 	if (len < 3)
 	    return;
 	port->linestate_mask = option[2];
+	outopt[0] = 44;
+	outopt[1] = 110;
+	outopt[2] = port->linestate_mask;
+	telnet_send_option(&port->tn_data, outopt, 3);
 	break;
 
     case 11: /* SET-MODEMSTATE-MASK */
 	if (len < 3)
 	    return;
 	port->modemstate_mask = option[2];
+	outopt[0] = 44;
+	outopt[1] = 111;
+	outopt[2] = port->modemstate_mask;
+	telnet_send_option(&port->tn_data, outopt, 3);
 	break;
 
     case 12: /* PURGE_DATA */
@@ -2279,6 +2374,10 @@ com_port_handler(void *cb_data, unsigned char *option, int len)
 	break;
     purge_found:
 	tcflush(port->devfd, val);
+	outopt[0] = 44;
+	outopt[1] = 112;
+	outopt[2] = option[2];
+	telnet_send_option(&port->tn_data, outopt, 3);
 	break;
 
     case 6: /* NOTIFY-LINESTATE */
