@@ -22,12 +22,20 @@
 
 #include "controller.h"
 
+#ifdef USE_UUCP_LOCKING
+extern int uucp_locking_enabled;
+#endif
+
 /* Create a port given the criteria. */
 char * portconfig(char *portnum,
 		  char *state,
 		  char *timeout,
 		  char *devname,
-		  char *devcfg);
+		  char *devcfg,
+		  int  config_num);
+
+/* Clear out any old ports on a reconfigure. */
+void clear_old_port_config(int config_num);
 
 /* Initialize the data transfer code. */
 void dataxfer_init(void);
@@ -37,6 +45,9 @@ void dataxfer_init(void);
    integers.  Error output will be generated on invalid data. */
 void showports(struct controller_info *cntlr, char *portspec);
 
+/* Show information about a port (as above) but in a one-line format. */
+void showshortports(struct controller_info *cntlr, char *portspec);
+
 /* Set the port's timeout.  The parameters are all strings that the
    routine will convert to integers.  Error output will be generated
    on invalid data. */
