@@ -53,6 +53,11 @@ selector_t *ser2net_sel;
 static char *help_string =
 "%s: Valid parameters are:\n"
 "  -c <config file> - use a config file besides /etc/ser2net.conf\n"
+"  -C <config line> - Handle a single configuration line.  This may be\n"
+"     specified multiple times for multiple lines.  This is just like a\n"
+"     line in the config file.  This disables the default config file,\n"
+"     you must specify a -c after the last -C to have it read a config\n"
+"     file, too."
 "  -p <controller port> - Start a controller session on the given TCP port\n"
 "  -n - Don't detach from the controlling terminal\n"
 "  -d - Don't detach and send debug I/O to standard output\n"
@@ -65,8 +70,10 @@ static char *help_string =
 void
 reread_config(void)
 {
-    syslog(LOG_INFO, "Got SIGHUP, re-reading configuration");
-    readconfig(config_file);
+    if (config_file) {
+	syslog(LOG_INFO, "Got SIGHUP, re-reading configuration");
+	readconfig(config_file);
+    }
 }
 
 void
@@ -82,6 +89,14 @@ main(int argc, char *argv[])
     int i;
     int err;
 
+    err = sel_alloc_selector(&ser2net_sel);
+    if (err) {
+	fprintf(stderr,
+		"Could not initialize ser2net selector: '%s'\n",
+		strerror(err));
+	return -1;
+    }
+
     for (i=1; i<argc; i++) {
 	if ((argv[i][0] != '-') || (strlen(argv[i]) != 2)) {
 	    fprintf(stderr, "Invalid argument: '%s'\n", argv[i]);
@@ -102,6 +117,17 @@ main(int argc, char *argv[])
 	    cisco_ios_baud_rates = 1;
 	    break;
 
+	case 'C':
+	    /* Get a config line. */
+	    i++;
+	    if (i == argc) {
+		fprintf(stderr, "No config line specified with -C\n");
+		arg_error(argv[0]);
+	    }
+	    handle_config_line(argv[i]);
+	    config_file = NULL;
+	    break;
+
 	case 'c':
 	    /* Get a config file. */
 	    i++;
@@ -138,13 +164,6 @@ main(int argc, char *argv[])
 	}
     }
 
-    err = sel_alloc_selector(&ser2net_sel);
-    if (err) {
-	fprintf(stderr,
-		"Could not initialize ser2net selector: '%s'\n",
-		strerror(err));
-	return -1;
-    }
     setup_sighup();
     if (config_port != NULL) {
 	if (controller_init(config_port) == -1) {
@@ -153,8 +172,10 @@ main(int argc, char *argv[])
 	}
     }
 
-    if (readconfig(config_file) == -1) {
-	return 1;
+    if (config_file) {
+	if (readconfig(config_file) == -1) {
+	    return 1;
+	}
     }
 
     if (detach) {
@@ -162,6 +183,7 @@ main(int argc, char *argv[])
 
 	/* Detach from the calling terminal. */
 	openlog("ser2net", LOG_PID | LOG_CONS, LOG_DAEMON);
+	syslog(LOG_NOTICE, "ser2net startup");
 	if ((pid = fork()) > 0) {
 	    exit(0);
 	} else if (pid < 0) {
